{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeFamilies    #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE LambdaCase #-}

module Gargantext.System.Logging (
    LogLevel(..)
  , HasLogger(..)
  , MonadLogger(..)
  , parseLogLevel
  , renderLogLevel
  , prop_loglevel_roundtrip
  , logM
  , logLocM
  , logLoc
  , withLogger
  , withLoggerIO
  ) where

import Control.Exception.Safe (MonadMask, bracket)
import Control.Monad (when)
import Control.Monad.IO.Class
import Control.Monad.Trans.Control
import Data.Kind (Type)
import Data.Text qualified as T
import Data.Time.Clock (getCurrentTime)
import Language.Haskell.TH hiding (Type)
import Language.Haskell.TH.Syntax qualified as TH
import Prelude
import System.Environment (lookupEnv)
import Text.Read (readMaybe)


data LogLevel =
  -- | Debug messages
    DEBUG
  -- | Information
  | INFO
  -- | Normal runtime conditions
  | WARNING
  -- | General Errors
  | ERROR
  deriving (Show, Eq, Ord, Enum, Bounded, Read)

renderLogLevel :: LogLevel -> T.Text
renderLogLevel = \case
  DEBUG   -> "debug"
  INFO    -> "info"
  WARNING -> "warning"
  ERROR   -> "error"

parseLogLevel :: T.Text -> Either T.Text LogLevel
parseLogLevel = \case
  "debug"   -> Right DEBUG
  "info"    -> Right INFO
  "warning" -> Right WARNING
  "warn"    -> Right WARNING
  "error"   -> Right ERROR
  xs -> Left ("Invalid log level found: " <> xs)

prop_loglevel_roundtrip :: LogLevel -> Bool
prop_loglevel_roundtrip ll = (parseLogLevel . renderLogLevel $ ll) == Right ll

-- | This is a barebore logging interface which we
-- can extend to plug a proper logging library, without
-- the details of the logger cropping up everywhere in
-- the rest of the codebase.
class HasLogger m where
  data family Logger m        :: Type
  type family LogInitParams m :: Type
  type family LogPayload m    :: Type
  initLogger    :: LogInitParams m -> (forall m1. MonadIO m1 => m1 (Logger m))
  destroyLogger :: Logger m        -> (forall m1. MonadIO m1 => m1 ())
  logMsg        :: Logger m        -> LogLevel -> LogPayload m -> m ()
  logTxt        :: Logger m        -> LogLevel -> T.Text -> m ()

-- | Separate typeclass to get hold of a 'Logger' from within a monad.
-- We keey 'HasLogger' and 'MonadLogger' separate to enforce compositionality,
-- i.e. we can still give instances to 'HasLogger' for things like 'IO' without
-- having to force actually acquiring a logger for those monads.
class HasLogger m => MonadLogger m where
  getLogger :: m (Logger m)

-- | A variant of 'logTxt' that doesn't require passing an explicit 'Logger'.
logM :: (Monad m, MonadLogger m) => LogLevel -> T.Text -> m ()
logM level msg = do
  logger <- getLogger
  logTxt logger level msg

-- | Like 'logM', but it automatically adds the file and line number to
-- the output log.
logLocM :: ExpQ
logLocM = [| \level msg ->
  let loc = $(getLocTH)
  in logM level (formatWithLoc loc msg)
  |]

logLoc :: ExpQ
logLoc = [| \logger level msg ->
  let loc = $(getLocTH)
  in logTxt logger level (formatWithLoc loc msg)
  |]

formatWithLoc :: Loc -> T.Text -> T.Text
formatWithLoc loc msg = "[" <> locationToText <> "] " <> msg
  where
    locationToText :: T.Text
    locationToText = T.pack $ (loc_filename loc) ++ ':' : (line loc) ++ ':' : (char loc)
      where
        line = show . fst . loc_start
        char = show . snd . loc_start

getLocTH :: ExpQ
getLocTH = [| $(location >>= liftLoc) |]

liftLoc :: Loc -> Q Exp
liftLoc (Loc a b c (d1, d2) (e1, e2)) = [|Loc
    $(TH.lift a)
    $(TH.lift b)
    $(TH.lift c)
    ($(TH.lift d1), $(TH.lift d2))
    ($(TH.lift e1), $(TH.lift e2))
    |]

-- | exception-safe combinator that creates and destroys a logger.
-- Think about it like a 'bracket' function from 'Control.Exception'.
withLogger :: (MonadBaseControl IO m, MonadIO m, HasLogger m, MonadMask m)
           => LogInitParams m
           -> (Logger m -> m a)
           -> m a
withLogger params = bracket (initLogger params) destroyLogger

-- | Like 'withLogger', but it allows creating a 'Logger' that can run in
-- a different monad from within an 'IO' action.
withLoggerIO :: (MonadBaseControl IO m, HasLogger m)
             => LogInitParams m
             -> (Logger m -> IO a)
             -> IO a
withLoggerIO params act = bracket (initLogger params) destroyLogger act

-- | A plain logger in the IO monad, waiting for more serious logging solutions like
-- the one described in https://gitlab.iscpif.fr/gargantext/haskell-gargantext/issues/229
instance HasLogger IO where
  data instance Logger IO        = IOLogger LogLevel
  type instance LogInitParams IO = ()
  type instance LogPayload IO    = String
  initLogger () = do
    mLvl <- liftIO $ lookupEnv "GGTX_LOG_LEVEL"
    let lvl = case mLvl of
          Nothing      -> INFO
          Just s       ->
            case readMaybe s of
              Nothing -> error $ "unknown log level " <> s
              Just lvl' -> lvl'
    pure $ IOLogger lvl
  destroyLogger _ = pure ()
  logMsg (IOLogger minLvl) lvl msg = do
    t <- getCurrentTime
    when (lvl >= minLvl) $ do
      let pfx = "[" <> show t <> "] [" <> show lvl <> "] "
      putStrLn $ pfx <> msg
  logTxt lgr lvl msg = logMsg lgr lvl (T.unpack msg)
